@@ -5,6 +5,7 @@ from __future__ import annotations
 
 import logging
 import os
+import pathlib
 import pickle
 import struct
 import sys
@@ -23,8 +24,8 @@ from velbusaio.channels import (
     Temperature,
     ThermostatChannel,
 )
-from velbusaio.const import PRIORITY_LOW
-from velbusaio.helpers import get_cache_dir, handle_match, keys_exists
+from velbusaio.const import CHANNEL_LIGHT_VALUE, CHANNEL_MEMO_TEXT, PRIORITY_LOW
+from velbusaio.helpers import handle_match, keys_exists
 from velbusaio.messages.blind_status import BlindStatusMessage, BlindStatusNgMessage
 from velbusaio.messages.channel_name_part1 import (
     ChannelNamePart1Message,
@@ -44,6 +45,7 @@ from velbusaio.messages.channel_name_part3 import (
 from velbusaio.messages.channel_name_request import ChannelNameRequestMessage
 from velbusaio.messages.clear_led import ClearLedMessage
 from velbusaio.messages.counter_status import CounterStatusMessage
+from velbusaio.messages.counter_status_request import CounterStatusRequestMessage
 from velbusaio.messages.dimmer_channel_status import DimmerChannelStatusMessage
 from velbusaio.messages.dimmer_status import DimmerStatusMessage
 from velbusaio.messages.fast_blinking_led import FastBlinkingLedMessage
@@ -58,7 +60,7 @@ from velbusaio.messages.module_subtype import ModuleSubTypeMessage
 from velbusaio.messages.module_type import ModuleTypeMessage
 from velbusaio.messages.push_button_status import PushButtonStatusMessage
 from velbusaio.messages.read_data_from_memory import ReadDataFromMemoryMessage
-from velbusaio.messages.relay_status import RelayStatusMessage
+from velbusaio.messages.relay_status import RelayStatusMessage, RelayStatusMessage2
 from velbusaio.messages.sensor_temperature import SensorTemperatureMessage
 from velbusaio.messages.set_led import SetLedMessage
 from velbusaio.messages.slider_status import SliderStatusMessage
@@ -81,6 +83,7 @@ class Module:
         memorymap=None,
         build_year=None,
         build_week=None,
+        cache_dir=None,
     ) -> None:
         self._address = module_address
         self._type = module_type
@@ -92,13 +95,13 @@ class Module:
         self.memory_map_version = memorymap
         self.build_year = build_year
         self.build_week = build_week
+        self._cache_dir = cache_dir
         self._is_loading = False
         self._channels = {}
         self.loaded = False
 
     def initialize(self, writer: type) -> None:
         self._log = logging.getLogger("velbus-module")
-        self._log.setLevel(logging.DEBUG)
         self._writer = writer
         for chan in self._channels.values():
             chan._writer = writer
@@ -113,9 +116,8 @@ class Module:
                         del self._channels[i]
 
     def _cache(self) -> None:
-        if not os.path.isdir(get_cache_dir()):
-            os.mkdir(get_cache_dir())
-        with open(f"{get_cache_dir()}/{self._address}.p", "wb") as fl:
+        cfile = pathlib.Path(f"{self._cache_dir}/{self._address}.p")
+        with cfile.open("wb") as fl:
             pickle.dump(self, fl)
 
     def __getstate__(self) -> dict:
@@ -174,10 +176,21 @@ class Module:
             self.build_week,
         )
 
+    def calc_channel_offset(self, address) -> int:
+        _channel_offset = 0
+        if self._address != address:
+            for _sub_addr_key, _sub_addr_val in self._sub_address.items():
+                if _sub_addr_val == address:
+                    _channel_offset = 8 * _sub_addr_key
+                    break
+        return _channel_offset
+
     async def on_message(self, message) -> None:
         """
         Process received message
         """
+        _channel_offset = self.calc_channel_offset(message.address)
+
         if isinstance(
             message,
             (
@@ -207,8 +220,15 @@ class Module:
             self._process_channel_name_message(3, message)
         elif isinstance(message, MemoryDataMessage):
             await self._process_memory_data_message(message)
-        elif isinstance(message, RelayStatusMessage):
-            await self._channels[message.channel].update({"on": message.is_on()})
+        elif isinstance(message, (RelayStatusMessage, RelayStatusMessage2)):
+            await self._channels[message.channel].update(
+                {
+                    "on": message.is_on(),
+                    "inhibit": message.is_inhibited(),
+                    "forced_on": message.is_forced_on(),
+                    "disabled": message.is_disabled(),
+                }
+            )
         elif isinstance(message, SensorTemperatureMessage):
             chan = self._translate_channel_name(self._data["TemperatureChannel"])
             await self._channels[chan].update(
@@ -222,34 +242,49 @@ class Module:
             # update the current temp
             chan = self._translate_channel_name(self._data["TemperatureChannel"])
             if chan in self._channels:
-                await self._channels[chan].update({"cur": message.current_temp})
-            # self._target = message.target_temp
-            # self._cmode = message.mode_str
-            # self._cstatus = message.status_str
+                await self._channels[chan].update(
+                    {
+                        "cur": message.current_temp,
+                        "target": message.target_temp,
+                        "cmode": message.mode_str,
+                        "cstatus": message.status_str,
+                    }
+                )
         elif isinstance(message, PushButtonStatusMessage):
-            for channel in message.closed:
-                channel = self._translate_channel_name(channel)
-                await self._channels[channel].update({"closed": True})
-            for channel in message.opened:
-                channel = self._translate_channel_name(channel)
-                await self._channels[channel].update({"closed": False})
-        elif isinstance(message, ModuleStatusMessage):
-            for channel in self._channels.keys():
-                channel = self._translate_channel_name(channel)
-                if channel in message.closed:
+            _update_buttons = False
+            for channel_types in self._data["Channels"]:
+                if keys_exists(self._data, "Channels", channel_types, "Type"):
+                    if self._data["Channels"][channel_types]["Type"] == "Button":
+                        _update_buttons = True
+                        break
+            if _update_buttons:
+                for channel_id in range(1, 9):
+                    channel = self._translate_channel_name(channel_id + _channel_offset)
+                    if channel_id in message.closed:
+                        await self._channels[channel].update({"closed": True})
+                    if channel_id in message.opened:
+                        await self._channels[channel].update({"closed": False})
+        elif isinstance(message, (ModuleStatusMessage)):
+            for channel_id in range(1, 9):
+                channel = self._translate_channel_name(channel_id + _channel_offset)
+                if channel_id in message.closed:
+                    await self._channels[channel].update({"closed": True})
+                elif channel in self._channels and isinstance(
+                    self._channels[channel], (Button, ButtonCounter)
+                ):
+                    await self._channels[channel].update({"closed": False})
+        elif isinstance(message, (ModuleStatusMessage2)):
+            for channel_id in range(1, 9):
+                channel = self._translate_channel_name(channel_id + _channel_offset)
+                if channel_id in message.closed:
                     await self._channels[channel].update({"closed": True})
                 elif isinstance(self._channels[channel], (Button, ButtonCounter)):
                     await self._channels[channel].update({"closed": False})
-        elif isinstance(message, ModuleStatusMessage2):
-            for channel in self._channels.keys():
-                channel = self._translate_channel_name(channel)
-                if channel in message.closed:
-                    await self._channels[channel].update({"closed": True})
-                elif isinstance(self._channels[channel], (Button, ButtonCounter)):
-                    await self._channels[channel].update({"closed": False})
-                if channel in message.enabled:
+                if channel_id in message.enabled:
                     await self._channels[channel].update({"enabled": True})
-                elif isinstance(self._channels[channel], (Button, ButtonCounter)):
+                elif channel in self._channels and isinstance(
+                    self._channels[channel], (Button, ButtonCounter)
+                ):
                     await self._channels[channel].update({"enabled": False})
         elif isinstance(message, CounterStatusMessage) and isinstance(
             self._channels[message.channel], ButtonCounter
@@ -263,51 +298,50 @@ class Module:
                 }
             )
         elif isinstance(message, ModuleStatusPirMessage):
-            await self._channels[99].update({"cur": message.light_value})
+            await self._channels[CHANNEL_LIGHT_VALUE].update(
+                {"cur": message.light_value}
+            )
         elif isinstance(message, UpdateLedStatusMessage):
-            for channel in self._channels.keys():
-                channel = self._translate_channel_name(message.channel)
-                if channel in message.led_slow_blinking:
+            for channel_id in range(1, 9):
+                channel = self._translate_channel_name(channel_id + _channel_offset)
+                if channel_id in message.led_slow_blinking:
                     await self._channels[channel].update({"led_state": "slow"})
-                if channel in message.led_fast_blinking:
+                if channel_id in message.led_fast_blinking:
                     await self._channels[channel].update({"led_state": "fast"})
-                if channel in message.led_on:
+                if channel_id in message.led_on:
                     await self._channels[channel].update({"led_state": "on"})
                 if (
-                    channel not in message.led_slow_blinking
-                    and channel not in message.led_fast_blinking
-                    and channel not in message.led_on
+                    channel_id not in message.led_slow_blinking
+                    and channel_id not in message.led_fast_blinking
+                    and channel_id not in message.led_on
                 ):
                     await self._channels[channel].update({"led_state": "off"})
         elif isinstance(message, SetLedMessage):
-            for channel in self._channels.keys():
-                if channel in message.leds:
-                    channel = self._translate_channel_name(channel)
+            for channel_id in range(1, 9):
+                channel = self._translate_channel_name(channel_id + _channel_offset)
+                if channel_id in message.leds:
                     await self._channels[channel].update({"led_state": "on"})
         elif isinstance(message, ClearLedMessage):
-            for channel in self._channels.keys():
-                if channel in message.leds:
-                    channel = self._translate_channel_name(channel)
+            for channel_id in range(1, 9):
+                channel = self._translate_channel_name(channel_id + _channel_offset)
+                if channel_id in message.leds:
                     await self._channels[channel].update({"led_state": "off"})
         elif isinstance(message, SlowBlinkingLedMessage):
-            for channel in self._channels.keys():
-                if channel in message.leds:
-                    channel = self._translate_channel_name(channel)
+            for channel_id in range(1, 9):
+                channel = self._translate_channel_name(channel_id + _channel_offset)
+                if channel_id in message.leds:
                     await self._channels[channel].update({"led_state": "slow"})
         elif isinstance(message, FastBlinkingLedMessage):
-            for channel in self._channels.keys():
-                if channel in message.leds:
-                    channel = self._translate_channel_name(channel)
+            for channel_id in range(1, 9):
+                channel = self._translate_channel_name(channel_id + _channel_offset)
+                if channel_id in message.leds:
                     await self._channels[channel].update({"led_state": "fast"})
-        elif isinstance(message, DimmerChannelStatusMessage):
+        elif isinstance(message, (DimmerChannelStatusMessage, DimmerStatusMessage)):
             channel = self._translate_channel_name(message.channel)
             await self._channels[channel].update({"state": message.cur_dimmer_state()})
         elif isinstance(message, SliderStatusMessage):
             channel = self._translate_channel_name(message.channel)
             await self._channels[channel].update({"state": message.cur_slider_state()})
-        elif isinstance(message, DimmerStatusMessage):
-            channel = self._translate_channel_name(message.channel)
-            await self._channels[channel].update({"state": message.cur_dimmer_state()})
         elif isinstance(message, BlindStatusNgMessage):
             channel = self._translate_channel_name(message.channel)
             await self._channels[channel].update(
@@ -325,7 +359,7 @@ class Module:
         """
         return self._channels
 
-    async def load(self) -> None:
+    async def load(self, from_cache=False) -> None:
         """
         Retrieve names of channels
         """
@@ -334,6 +368,8 @@ class Module:
         # as the submodule address maps to the main module
         # this method can be called multiple times
         if self._is_loading or self.loaded:
+            if from_cache:
+                await self._request_module_status()
             return
         self._log.info("Load Module")
         # start the loading
@@ -367,6 +403,11 @@ class Module:
             return 0
         return max(self._channels.keys())
 
+    async def set_memo_text(self, txt: str) -> None:
+        if CHANNEL_MEMO_TEXT not in self._channels.keys():
+            return
+        await self._channels[CHANNEL_MEMO_TEXT].set(txt)
+
     async def _process_memory_data_message(self, message) -> None:
         addr = "{high:02X}{low:02X}".format(
             high=message.high_address, low=message.low_address
@@ -432,10 +473,22 @@ class Module:
         return True
 
     async def _request_module_status(self) -> None:
-        # request the module status (if available for this module
-        msg = ModuleStatusRequestMessage(self._address)
-        msg.channels = list(range(1, 9))
-        await self._writer(msg)
+        """Request current state of channels."""
+        if "Channels" not in self._data:
+            # some modules have no channels
+            return
+        mod_stat_req_msg = ModuleStatusRequestMessage(self._address)
+        counter_msg = None
+        for chan, chan_data in self._data["Channels"].items():
+            if int(chan) < 9 and chan_data["Type"] in ("Blind", "Dimmer", "Relay"):
+                mod_stat_req_msg.channels.append(int(chan))
+            if chan_data["Type"] == "ButtonCounter":
+                if counter_msg is None:
+                    counter_msg = CounterStatusRequestMessage(self._address)
+                counter_msg.channels.append(int(chan))
+        await self._writer(mod_stat_req_msg)
+        if counter_msg is not None:
+            await self._writer(counter_msg)
 
     async def _request_channel_name(self) -> None:
         # request the module channel names
